@@ -1,10 +1,13 @@
 import json
 from pathlib import Path
+from stackl_client.models import StackStage
 from sys import exit
 
 import click
 import stackl_client
 import yaml
+from urllib3.exceptions import NewConnectionError, MaxRetryError
+from stackl_client.exceptions import ApiException, ApiValueError
 from commands.autocomplete import show_progress_bar
 from context import StacklContext
 from mergedeep import merge
@@ -16,7 +19,7 @@ from mergedeep import merge
 @click.option('-p', '--params', default=[], multiple=True)
 @click.option('-t', '--tags', default="{}")
 @click.option('-r', '--replicas', default="{}")
-@click.option('-s', '--secrets', default="{}")
+@click.option('--secrets', default="{}")
 @click.option('-e', '--service-params', default="{}")
 @click.option('--service-secrets', default="{}")
 @click.option('--services', default=[])
@@ -24,14 +27,25 @@ from mergedeep import merge
 @click.argument('instance-name', required=False)
 def apply(directory, config_file, params, tags, secrets, service_params,
           service_secrets, replicas, services, instance_name, show_progress):
-    stackl_context = StacklContext()
-    if instance_name is None:
-        upload_files(directory, stackl_context)
-    else:
-        apply_stack_instance(config_file, params, tags, secrets,
-                             service_params, service_secrets, replicas,
-                             services, stackl_context, instance_name,
-                             show_progress)
+    try:
+        stackl_context = StacklContext()
+        if instance_name is None:
+            upload_files(directory, stackl_context)
+        else:
+            apply_stack_instance(config_file, params, tags, secrets,
+                                service_params, service_secrets, replicas,
+                                services, stackl_context, instance_name,
+                                show_progress)
+    except ApiException as e:
+        click.echo(e.body)
+        exit(1)
+    except ApiValueError as e:
+        click.echo(e.body)
+        exit(1)
+    except (NewConnectionError, MaxRetryError) as e:
+        click.echo("Unable to connect to Stackl host")
+        click.echo(e)
+        exit(1)
 
 
 def apply_stack_instance(config_file, params, tags, secrets, service_params,
@@ -44,6 +58,7 @@ def apply_stack_instance(config_file, params, tags, secrets, service_params,
     final_params = {**config_doc['params'], **final_params}
     tags = json.loads(tags)
     replicas = json.loads(replicas)
+    stages = []
     if "replicas" in config_doc:
         replicas = {**config_doc['replicas'], **replicas}
     secrets = json.loads(secrets)
@@ -59,7 +74,12 @@ def apply_stack_instance(config_file, params, tags, secrets, service_params,
         tags = {**config_doc['tags'], **tags}
     if "services" in config_doc:
         services = config_doc['services']
-    invocation = stackl_client.StackInstanceInvocation(
+    if "stages" in config_doc:
+        stages_list = config_doc['stages']
+        for stage in stages_list:
+            stage_obj = StackStage(name=stage['name'], services=stage['services'])
+            stages.append(stage_obj)
+    invocation = stackl_client.models.StackInstanceInvocation(
         stack_instance_name=instance_name,
         stack_infrastructure_template=config_doc[
             "stack_infrastructure_template"],
@@ -70,13 +90,50 @@ def apply_stack_instance(config_file, params, tags, secrets, service_params,
         service_secrets=service_secrets,
         secrets=secrets,
         services=services,
+        stages=stages,
         tags=tags)
     try:
+        invocation = stackl_client.models.StackInstanceUpdate(
+        stack_instance_name=instance_name,
+        params=final_params,
+        replicas=replicas,
+        service_params=service_params,
+        service_secrets=service_secrets,
+        secrets=secrets,
+        services=services,
+        stages=stages,
+        tags=tags)
         stackl_context.stack_instances_api.get_stack_instance(instance_name)
         res = stackl_context.stack_instances_api.put_stack_instance(invocation)
-    except stackl_client.exceptions.ApiException:
-        res = stackl_context.stack_instances_api.post_stack_instance(
-            invocation)
+    except stackl_client.exceptions.NotFoundException as e:
+        try:
+            invocation = stackl_client.models.StackInstanceInvocation(
+            stack_instance_name=instance_name,
+            stack_infrastructure_template=config_doc[
+                "stack_infrastructure_template"],
+            stack_application_template=config_doc["stack_application_template"],
+            params=final_params,
+            replicas=replicas,
+            service_params=service_params,
+            service_secrets=service_secrets,
+            secrets=secrets,
+            services=services,
+            stages=stages,
+            tags=tags)
+            res = stackl_context.stack_instances_api.post_stack_instance(invocation)
+        except ApiException as e:
+            click.echo(e)
+            exit(1)
+    except ApiException as e:
+        click.echo(e.body)
+        exit(1)
+    except ApiValueError as e:
+        click.echo(e.body)
+        exit(1)
+    except (NewConnectionError, MaxRetryError) as e:
+        click.echo("Unable to connect to Stackl host")
+        click.echo(e)
+        exit(1)
 
     click.echo(res)
 
@@ -110,11 +167,18 @@ def upload_file(stackl_doc, stackl_context, path):
         click.echo(
             f"Succesfully applied {stackl_doc['name']} with type {stackl_doc['type']}"
         )
-    except stackl_client.exceptions.ApiException as e:
+    except ApiException as e:
         click.echo(
             f"Failed to apply {stackl_doc['name']} with type {stackl_doc['type']}: {e.body}"
         )
         exit(1)
+    except ApiValueError as e:
+        click.echo(e.body)
+        exit(1)
+    except (NewConnectionError, MaxRetryError) as e:
+        click.echo("Unable to connect to Stackl host")
+        click.echo(e)
+        exit(1)
 
 
 def upload_files(directory, stackl_context):
